from unittest import skip

from cms.models import PlaceholderRelationField
from cms.test_utils.testcases import CMSTestCase
from cms.toolbar.items import TemplateItem
from cms.toolbar.utils import get_object_preview_url
from cms.utils import get_current_site
from django.contrib import admin
from django.contrib.auth.models import Permission
from django.core import mail
from django.template.loader import render_to_string
from django.test import RequestFactory, override_settings
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from djangocms_versioning import (
    admin as versioning_admin,
    conf,
    models as versioning_models,
)
from djangocms_versioning.cms_config import VersioningCMSConfig
from djangocms_versioning.constants import ARCHIVED, DRAFT, PUBLISHED, UNPUBLISHED
from djangocms_versioning.emails import get_full_url
from djangocms_versioning.helpers import (
    create_version_lock,
    placeholder_content_is_unlocked_for_user,
    version_list_url,
)
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils import factories
from djangocms_versioning.test_utils.blogpost.models import BlogPost
from djangocms_versioning.test_utils.factories import (
    FancyPollFactory,
    PageVersionFactory,
    PlaceholderFactory,
    UserFactory,
)
from djangocms_versioning.test_utils.polls.cms_config import PollsCMSConfig
from djangocms_versioning.test_utils.test_helpers import (
    find_toolbar_buttons,
    get_toolbar,
    toolbar_button_exists,
)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class AdminLockedFieldTestCase(CMSTestCase):

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

        site = admin.AdminSite()
        self.hijacked_admin = versioning_admin.VersionAdmin(Version, site)

    def test_version_admin_contains_locked_field(self):
        """
        The locked column exists in the admin field list
        """
        request = RequestFactory().get("/admin/djangocms_versioning/pollcontentversion/")
        self.assertIn(_("locked"), self.hijacked_admin.get_list_display(request))

    def test_version_lock_state_locked(self):
        """
        A published version does not have an entry in the locked column in the admin
        """
        published_version = factories.PollVersionFactory(state=PUBLISHED)

        self.assertEqual("", self.hijacked_admin.locked(published_version))

    def test_version_lock_state_unlocked(self):
        """
        A locked draft version does have an entry in the locked column in the version
        admin and is not empty
        """
        draft_version = factories.PollVersionFactory(state=DRAFT)
        create_version_lock(draft_version, self.get_superuser())

        self.assertNotEqual("", self.hijacked_admin.locked(draft_version))


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class AdminPermissionTestCase(CMSTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.versionable = PollsCMSConfig.versioning[0]

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

        self.superuser = self.get_superuser()
        self.user_has_change_perms = self._create_user(
            "user_has_unlock_perms",
            is_staff=True,
            permissions=["change_pollcontentversion", "delete_versionlock"],
        )

    def test_user_has_change_permission(self):
        """
        The user who created the version has permission to change it
        """
        version = factories.PollVersionFactory(
            state=DRAFT,
            created_by=self.user_has_change_perms,
            locked_by=self.user_has_change_perms,
        )
        url = self.get_admin_url(self.versionable.content_model, "change", version.content.pk)

        with self.login_user_context(self.user_has_change_perms):
            response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_user_does_not_have_change_permission(self):
        """
        A different user from the user who created
        the version does not have permission to change it
        """
        author = factories.UserFactory(is_staff=True)
        version = factories.PollVersionFactory(state=DRAFT, created_by=author, locked_by=author)

        url = self.get_admin_url(self.versionable.content_model, "change", version.content.pk)
        with self.login_user_context(self.user_has_change_perms):
            response = self.client.get(url)

        self.assertIsNotNone(version.locked_by)  # Was locked
        self.assertEqual(response.status_code, 403)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class VersionLockUnlockTestCase(CMSTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.versionable = PollsCMSConfig.versioning[0]
        cls.default_permissions = ["change_pollcontentversion"]

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

        self.superuser = self.get_superuser()
        self.user_author = self._create_user(
            "author",
            is_staff=True,
            permissions=self.default_permissions,
        )
        self.user_has_no_unlock_perms = self._create_user(
            "user_has_no_unlock_perms",
            is_staff=True,
            permissions=self.default_permissions,
        )
        self.user_has_unlock_perms = self._create_user(
            "user_has_unlock_perms",
            is_staff=True,
            permissions=["delete_versionlock"] + self.default_permissions,
        )

    def test_unlock_view_refuses_get(self):
        poll_version = factories.PollVersionFactory(
            state=PUBLISHED,
            created_by=self.superuser,
            locked_by=self.superuser,
        )
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)

        # 404 when not in draft
        with self.login_user_context(self.superuser):
            response = self.client.get(unlock_url, follow=True)

        self.assertEqual(response.status_code, 405)

    def test_unlock_view_redirects_to_admin_dashboard_for_non_existent_id(self):
        poll_version = factories.PollVersionFactory(
            state=PUBLISHED,
            created_by=self.superuser,
            locked_by=self.superuser,
        )
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock",
                                        poll_version.pk+314159)

        # 404 when not in draft
        with self.login_user_context(self.superuser):
            response = self.client.post(unlock_url, follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "dashboard")

    def test_unlock_view_redirects_404_when_not_draft(self):
        poll_version = factories.PollVersionFactory(
            state=PUBLISHED,
            created_by=self.superuser,
            locked_by=self.superuser,
        )
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)

        # 404 when not in draft
        with self.login_user_context(self.superuser):
            response = self.client.post(unlock_url, follow=True)

        self.assertEqual(response.status_code, 404)

    def test_unlock_view_not_possible_for_user_with_no_permissions(self):
        poll_version = factories.PollVersionFactory(
            state=DRAFT,
            created_by=self.user_author,
            locked_by=self.user_author,
        )
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)

        with self.login_user_context(self.user_has_no_unlock_perms):
            response = self.client.post(unlock_url, follow=True)

        self.assertEqual(response.status_code, 403)

        # Fetch the latest state of this version
        updated_poll_version = Version.objects.get(pk=poll_version.pk)

        # The version is still locked
        self.assertIsNotNone(updated_poll_version.locked_by)
        # The author is unchanged
        self.assertEqual(updated_poll_version.locked_by, self.user_author)

    def test_unlock_view_possible_for_user_with_permissions(self):
        poll_version = factories.PollVersionFactory(
            state=DRAFT,
            created_by=self.user_author,
            locked_by=self.user_author
        )
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)

        with self.login_user_context(self.user_has_unlock_perms):
            response = self.client.post(unlock_url, follow=True)

        self.assertEqual(response.status_code, 200)

        # Fetch the latest state of this version
        updated_poll_version = Version.objects.get(pk=poll_version.pk)

        # The version is not locked
        self.assertFalse(hasattr(updated_poll_version, "versionlock"))

    @skip("Requires clarification if this is still a valid requirement!")
    def test_unlock_link_not_present_for_author(self):
        # FIXME: May be redundant now as this requirement was probably removed at a later date due
        #  to the fact that an author may be asked to unlock their version for someone else to use!
        author = self.get_superuser()
        poll_version = factories.PollVersionFactory(state=DRAFT, created_by=author, locked_by=author)
        changelist_url = version_list_url(poll_version.content)
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)
        unlock_control = render_to_string(
            "djangocms_version_locking/admin/unlock_icon.html",
            {"unlock_url": unlock_url}
        )

        with self.login_user_context(author):
            response = self.client.get(changelist_url)

        self.assertNotContains(response, unlock_control, html=True)

    def test_unlock_link_not_present_for_user_with_no_unlock_privileges(self):
        poll_version = factories.PollVersionFactory(
            state=DRAFT,
            created_by=self.user_author,
            locked_by=self.user_author)
        changelist_url = version_list_url(poll_version.content)
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)
        exprected_disabled_button = (
            f'<a class="btn cms-form-post-method cms-action-btn cms-action-unlock js-action js-keep-sideframe" '
            f'href="{unlock_url}" title="Unlock"><span class="cms-icon cms-icon-unlock"></span></a>'
        )
        with self.login_user_context(self.user_has_no_unlock_perms):
            response = self.client.post(changelist_url)
        self.assertInHTML(exprected_disabled_button, response.content.decode("utf-8"))

    def test_unlock_link_present_for_user_with_privileges(self):
        poll_version = factories.PollVersionFactory(
            state=DRAFT,
            created_by=self.user_author,
            locked_by=self.user_author,
        )
        changelist_url = version_list_url(poll_version.content)
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", poll_version.pk)
        unlock_control = "cms-action-unlock"

        with self.login_user_context(self.user_has_unlock_perms):
            response = self.client.post(changelist_url)

        self.assertContains(response, unlock_control)  # Action button present
        self.assertContains(response, unlock_url)  # Not present for disabled action button

    def test_unlock_link_only_present_for_draft_versions(self):
        draft_version = factories.PollVersionFactory(created_by=self.user_author, locked_by=self.user_author)
        published_version = Version.objects.create(
            content=factories.PollContentFactory(poll=draft_version.content.poll),
            created_by=factories.UserFactory(),
            state=PUBLISHED
        )
        draft_unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", draft_version.pk)
        draft_unlock_control = "cms-action-unlock"
        published_unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", published_version.pk)
        published_unlock_control = "cms-action-unlock"
        changelist_url = version_list_url(draft_version.content)

        with self.login_user_context(self.superuser):
            response = self.client.post(changelist_url)

        # The draft version unlock control exists
        self.assertContains(response, draft_unlock_control)
        self.assertContains(response, draft_unlock_url)
        # The published version exists
        self.assertContains(response, published_unlock_control)
        self.assertNotContains(response, published_unlock_url)

    def test_unlock_and_new_user_edit_creates_version_lock(self):
        """
        When a version is unlocked a different user (or the same) can then visit the edit link and take
        ownership of the version, this creates a version lock for the editing user
        """
        draft_version = factories.PollVersionFactory(created_by=self.user_author, locked_by=self.user_author)
        draft_unlock_url = self.get_admin_url(self.versionable.version_model_proxy,
                                              "unlock", draft_version.pk)

        # The version is owned by the author
        self.assertEqual(draft_version.created_by, self.user_author)
        # The version lock exists and is owned by the author
        self.assertEqual(draft_version.locked_by, self.user_author)

        # Unlock the version with a different user with unlock permissions
        with self.login_user_context(self.user_has_unlock_perms):
            self.client.post(draft_unlock_url, follow=True)

        updated_draft_version = Version.objects.get(pk=draft_version.pk)
        updated_draft_edit_url = self.get_admin_url(
            self.versionable.version_model_proxy,
            "edit_redirect", updated_draft_version.pk
        )

        # The version is still owned by the author
        self.assertEqual(updated_draft_version.created_by, self.user_author)
        # The version lock does not exist
        self.assertIsNone(updated_draft_version.locked_by)

        # Visit the edit page with a user without unlock permissions
        with self.login_user_context(self.user_has_no_unlock_perms):
            self.client.post(updated_draft_edit_url)

        updated_draft_version = Version.objects.get(pk=draft_version.pk)

        # The version is still owned by the author
        self.assertEqual(updated_draft_version.created_by, self.user_author)
        # The version lock exists and is now owned by the user with no permissions
        self.assertEqual(updated_draft_version.locked_by, self.user_has_no_unlock_perms)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class VersionLockEditActionStateTestCase(CMSTestCase):

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

        self.superuser = self.get_superuser()
        self.user_author = self._create_user("author", is_staff=True, is_superuser=False)
        self.versionable = PollsCMSConfig.versioning[0]
        self.version_admin = admin.site._registry[self.versionable.version_model_proxy]

    def test_edit_action_link_enabled_state(self):
        """
        The edit action is active
        """
        version = factories.PollVersionFactory(created_by=self.user_author, locked_by=self.user_author)
        author_request = RequestFactory()
        author_request.user = self.user_author
        otheruser_request = RequestFactory()
        otheruser_request.user = self.superuser

        actual_enabled_state = self.version_admin._get_edit_link(version, author_request)

        self.assertNotIn("inactive", actual_enabled_state)

    def test_edit_action_link_disabled_state(self):
        """
        The edit action is disabled for a different user to the locked user
        """
        version = factories.PollVersionFactory(created_by=self.user_author, locked_by=self.user_author)
        author_request = RequestFactory()
        author_request.user = self.user_author
        otheruser_request = RequestFactory()
        otheruser_request.user = self.superuser
        expected_disabled_state = ""

        actual_disabled_state = self.version_admin._get_edit_link(version, otheruser_request)

        self.assertFalse(version.check_edit_redirect.as_bool(self.superuser))
        self.assertEqual(expected_disabled_state, actual_disabled_state)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class VersionLockEditActionSideFrameTestCase(CMSTestCase):
    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

        self.superuser = self.get_superuser()
        self.user_author = self._create_user("author", is_staff=True, is_superuser=False)
        self.versionable = PollsCMSConfig.versioning[0]
        self.version_admin = admin.site._registry[self.versionable.version_model_proxy]

    def test_version_unlock_keep_side_frame(self):
        """
        When clicking on an versionables enabled unlock icon, the sideframe is kept open
        """
        version = factories.PollVersionFactory(created_by=self.user_author, locked_by=self.user_author)
        author_request = RequestFactory()
        author_request.user = self.user_author
        otheruser_request = RequestFactory()
        otheruser_request.user = self.superuser

        actual_enabled_state = self.version_admin._get_unlock_link(version, otheruser_request)

        # The url link should keep the sideframe open
        self.assertIn("js-keep-sideframe", actual_enabled_state)
        self.assertNotIn("js-close-sideframe", actual_enabled_state)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class VersionLockIndicatorTestCase(CMSTestCase):

    def setUp(self) -> None:
        self.LOCK_VERSIONS = conf.LOCK_VERSIONS
        conf.LOCK_VERSIONS = True

        self.superuser = self.get_superuser()
        self.user_author = self._create_user("author", is_staff=True, is_superuser=False)
        self.version_admin = admin.site._registry[BlogPost]

    def tearDown(self) -> None:
        conf.LOCK_VERSIONS = self.LOCK_VERSIONS

    def test_unlock_action_in_indicator_menu(self):
        """The indicator drop down menu contains an entry to unlock a draft."""
        changelist_url = reverse("admin:blogpost_blogpost_changelist")
        version = factories.BlogPostVersionFactory(created_by=self.user_author, locked_by=self.user_author)
        expected_unlock_url = reverse("admin:djangocms_versioning_blogcontentversion_unlock", args=(version.pk,))

        with self.login_user_context(self.superuser):
            response = self.client.get(changelist_url)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "cms-icon cms-icon-unlock")
        self.assertContains(response, expected_unlock_url)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class CheckLockTestCase(CMSTestCase):

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

    def test_check_no_lock(self):
        user = self.get_superuser()
        version = PageVersionFactory(state=ARCHIVED)
        placeholder = PlaceholderFactory(source=version.content)

        self.assertTrue(placeholder_content_is_unlocked_for_user(placeholder, user))

    def test_check_locked_for_the_same_user(self):
        user = self.get_superuser()
        version = PageVersionFactory(created_by=user, locked_by=user)
        placeholder = PlaceholderFactory(source=version.content)

        self.assertTrue(placeholder_content_is_unlocked_for_user(placeholder, user))

    def test_check_locked_for_the_other_user(self):
        user1 = self.get_superuser()
        user2 = self.get_standard_user()
        version = PageVersionFactory(created_by=user1, locked_by=user1)
        placeholder = PlaceholderFactory(source=version.content)

        self.assertFalse(placeholder_content_is_unlocked_for_user(placeholder, user2))

    def test_check_no_lock_for_unversioned_model(self):
        user2 = self.get_standard_user()
        placeholder = PlaceholderFactory(source=FancyPollFactory())

        self.assertTrue(placeholder_content_is_unlocked_for_user(placeholder, user2))


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class CheckInjectTestCase(CMSTestCase):

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

    @skip("This test would require reloading of the django app configs.")
    def test_lock_check_is_injected_into_default_checks(self):
        self.assertIn(
            placeholder_content_is_unlocked_for_user,
            PlaceholderRelationField.default_checks,
        )


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class VersionLockNotificationEmailsTestCase(CMSTestCase):

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)

        self.superuser = self.get_superuser()
        self.user_author = self._create_user("author", is_staff=True, is_superuser=False)
        self.user_has_no_perms = self._create_user("user_has_no_perms", is_staff=True, is_superuser=False)
        self.user_has_unlock_perms = self._create_user("user_has_unlock_perms", is_staff=True, is_superuser=False)
        self.versionable = VersioningCMSConfig.versioning[0]

        # Set permissions
        delete_permission = Permission.objects.get(codename="delete_versionlock")
        self.user_has_unlock_perms.user_permissions.add(delete_permission)

    def test_notify_version_author_version_unlocked_email_sent_for_different_user(self):
        """
        The user unlocking a version that is authored buy a different user
        should be sent a notification email
        """
        draft_version = factories.PageVersionFactory(content__template="", created_by=self.user_author)
        draft_unlock_url = self.get_admin_url(self.versionable.version_model_proxy,
                                              "unlock", draft_version.pk)

        # Check that no emails exist
        self.assertEqual(len(mail.outbox), 0)

        # Unlock the version with a different user with unlock permissions
        with self.login_user_context(self.user_has_unlock_perms):
            self.client.post(draft_unlock_url, follow=True)

        site = get_current_site()
        expected_subject = "[Django CMS] ({site_name}) {title} - {description}".format(
            site_name=site.name,
            title=draft_version.content,
            description=_("Unlocked"),
        )
        expected_body = f"The following draft version has been unlocked by {self.user_has_unlock_perms} for their use."
        expected_version_url = get_full_url(
            get_object_preview_url(draft_version.content)
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, expected_subject)
        self.assertEqual(mail.outbox[0].to[0], self.user_author.email)
        self.assertIn(expected_body, mail.outbox[0].body)
        self.assertIn(expected_version_url, mail.outbox[0].body)

    def test_notify_version_author_version_unlocked_email_not_sent_for_different_user(self):
        """
        The user unlocking a version that authored the version should not be
        sent a notification email
        """
        draft_version = factories.PageVersionFactory(content__template="", created_by=self.user_author)
        draft_unlock_url = self.get_admin_url(self.versionable.version_model_proxy,
                                              "unlock", draft_version.pk)

        # Check that no emails exist
        self.assertEqual(len(mail.outbox), 0)

        # Unlock the version the same user who authored it
        with self.login_user_context(self.user_author):
            self.client.post(draft_unlock_url, follow=True)

        # Check that no emails still exist
        self.assertEqual(len(mail.outbox), 0)

    def test_notify_version_author_version_unlocked_email_contents_users_full_name_used(self):
        """
        The email contains the full name of the author
        """
        user = self.user_has_unlock_perms
        user.first_name = "Firstname"
        user.last_name = "Lastname"
        user.save()
        draft_version = factories.PageVersionFactory(content__template="", created_by=self.user_author)
        draft_unlock_url = self.get_admin_url(self.versionable.version_model_proxy,
                                              "unlock", draft_version.pk)

        # Check that no emails exist
        self.assertEqual(len(mail.outbox), 0)

        # Unlock the version with a different user with unlock permissions
        with self.login_user_context(user):
            self.client.post(draft_unlock_url, follow=True)

        expected_body = f"The following draft version has been unlocked by {user.get_full_name()} for their use."

        self.assertEqual(len(mail.outbox), 1)
        self.assertIn(expected_body, mail.outbox[0].body)

    def test_notify_version_author_version_unlocked_email_contents_users_username_used(self):
        """
        The email contains the  username of the author because no name is available
        """
        user = self.user_has_unlock_perms
        draft_version = factories.PageVersionFactory(content__template="", created_by=self.user_author)
        draft_unlock_url = self.get_admin_url(self.versionable.version_model_proxy,
                                              "unlock", draft_version.pk)

        # Check that no emails exist
        self.assertEqual(len(mail.outbox), 0)

        # Unlock the version with a different user with unlock permissions
        with self.login_user_context(user):
            self.client.post(draft_unlock_url, follow=True)

        expected_body = f"The following draft version has been unlocked by {user.username} for their use."

        self.assertEqual(len(mail.outbox), 1)
        self.assertIn(expected_body, mail.outbox[0].body)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class TestVersionsLockTestCase(CMSTestCase):

    def setUp(self):
        import importlib
        importlib.reload(conf)
        importlib.reload(versioning_admin)
        self.versionable = PollsCMSConfig.versioning[0]
        self.user = self.get_standard_user()

    def test_version_is_locked_for_draft(self):
        """
        A version lock is present when a content version is created in a draft state with a locked_by user
        """
        draft_version = factories.PollVersionFactory(state=DRAFT, created_by=self.user, locked_by=self.user)

        self.assertIsNotNone(draft_version.locked_by)

    def test_version_is_unlocked_for_publishing(self):
        """
        A version lock is not present when a content version is in a published or unpublished state
        """
        user = self.get_superuser()
        poll_version = factories.PollVersionFactory(state=DRAFT, created_by=user, locked_by=user)
        publish_url = self.get_admin_url(self.versionable.version_model_proxy, "publish", poll_version.pk)
        unpublish_url = self.get_admin_url(self.versionable.version_model_proxy, "unpublish", poll_version.pk)

        with self.login_user_context(user):
            self.client.post(publish_url)

        updated_poll_version = Version.objects.get(pk=poll_version.pk)

        # The state is now PUBLISHED
        self.assertEqual(updated_poll_version.state, PUBLISHED)
        # Version lock does not exist
        self.assertIsNone(updated_poll_version.locked_by)

        with self.login_user_context(user):
            self.client.post(unpublish_url)

        updated_poll_version = Version.objects.get(pk=poll_version.pk)

        # The state is now UNPUBLISHED
        self.assertEqual(updated_poll_version.state, UNPUBLISHED)
        # Version lock does not exist
        self.assertFalse(hasattr(updated_poll_version, "versionlock"))

    def test_version_is_unlocked_for_archived(self):
        """
        A version lock is not present when a content version is in an archived state
        """
        user = self.get_superuser()
        poll_version = factories.PollVersionFactory(state=DRAFT, created_by=user, locked_by=user)
        archive_url = self.get_admin_url(self.versionable.version_model_proxy, "archive", poll_version.pk)

        with self.login_user_context(user):
            self.client.post(archive_url)

        updated_poll_version = Version.objects.get(pk=poll_version.pk)

        # The state is now ARCHIVED
        self.assertEqual(updated_poll_version.state, ARCHIVED)
        # Version lock does not exist
        self.assertFalse(hasattr(updated_poll_version, "versionlock"))


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class TestVersionCopyLocks(CMSTestCase):

    def setUp(self) -> None:
        self.LOCK_VERSIONS = versioning_models.LOCK_VERSIONS
        versioning_models.LOCK_VERSIONS = True

    def tearDown(self) -> None:
        versioning_models.LOCK_VERSIONS = self.LOCK_VERSIONS

    def test_draft_version_copy_creates_draft_lock(self):
        """
        A version lock is created for a new draft version copied from a draft version
        """
        user = factories.UserFactory()
        draft_version = factories.PollVersionFactory(state=DRAFT)
        new_version = draft_version.copy(user)

        self.assertIsNotNone(new_version.locked_by)

    def test_published_version_copy_creates_draft_lock(self):
        """
        A version lock is created for a published version copied from a draft version
        """
        user = factories.UserFactory()
        published_version = factories.PollVersionFactory(state=PUBLISHED, locked_by=None)
        new_version = published_version.copy(user)

        self.assertIsNotNone(new_version.locked_by)

    def test_version_copy_adds_correct_locked_user(self):
        """
        A copied version creates a lock for the user that copied the version.
        The users should not be the same.
        """
        original_user = factories.UserFactory()
        original_version = factories.PollVersionFactory(created_by=original_user, locked_by=original_user)
        copy_user = factories.UserFactory()
        copied_version = original_version.copy(copy_user)

        self.assertNotEqual(original_user, copy_user)
        self.assertEqual(original_version.locked_by, original_user)
        self.assertEqual(copied_version.locked_by, copy_user)


@override_settings(DJANGOCMS_VERSIONING_LOCK_VERSIONS=True)
class VersionToolbarOverrideTestCase(CMSTestCase):

    def setUp(self) -> None:
        from cms.models.permissionmodels import GlobalPagePermission

        from djangocms_versioning import cms_toolbars
        self.LOCK_VERSIONS = cms_toolbars.LOCK_VERSIONS
        cms_toolbars.LOCK_VERSIONS = True

        self.user_has_change_perms = self._create_user(
            "user_default_perms",
            is_staff=True,
            permissions=["change_page", "add_page", "delete_page"],
        )
        # Grant permission (or Unlock button will not be shown)
        GlobalPagePermission.objects.create(
            user=self.user_has_change_perms,
        )

    def tearDown(self) -> None:
        from djangocms_versioning import cms_toolbars
        cms_toolbars.LOCK_VERSIONS = self.LOCK_VERSIONS

    def test_not_render_edit_button_when_not_content_mode(self):
        user = self.get_superuser()
        version = PageVersionFactory(created_by=user)

        toolbar = get_toolbar(version.content, user, edit_mode=True)
        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Edit", toolbar.toolbar))

    def test_no_edit_button_when_content_is_locked(self):
        user = self.get_superuser()
        user_2 = UserFactory(
            is_staff=True,
            is_superuser=True,
            username="admin2",
            email="admin2@123.com",
        )
        version = PageVersionFactory(created_by=user, locked_by=user)

        toolbar = get_toolbar(version.content, user_2, content_mode=True)
        toolbar.post_template_populate()
        edit_buttons = find_toolbar_buttons("Edit", toolbar.toolbar)
        self.assertListEqual(edit_buttons, [])

    def test_disabled_unlock_button_when_content_is_locked(self):
        user = self.get_superuser()
        user_2 = self.user_has_change_perms
        version = PageVersionFactory(created_by=user, locked_by=user)

        toolbar = get_toolbar(version.content, user_2, content_mode=True)
        toolbar.post_template_populate()

        unlock_buttons = find_toolbar_buttons("Unlock", toolbar.toolbar)
        self.assertEqual(len(unlock_buttons), 1)
        self.assertEqual(unlock_buttons[0].url, "#")  # disabled

    def test_enabled_unlock_button_when_content_is_locked(self):
        user = UserFactory(
            is_staff=True,
            is_superuser=True,
            username="admin2",
            email="admin2@123.com",
        )
        version = PageVersionFactory(created_by=user, locked_by=user)
        toolbar = get_toolbar(version.content, user=self.get_superuser(), content_mode=True)
        proxy_model = toolbar._get_proxy_model()
        expected_unlock_url = reverse(
            f"admin:{proxy_model._meta.app_label}_{proxy_model.__name__.lower()}_unlock",
            args=(version.pk,),
        )
        toolbar.post_template_populate()
        unlock_buttons = find_toolbar_buttons("Unlock", toolbar.toolbar)
        self.assertEqual(unlock_buttons[0].url, expected_unlock_url)  # enabled

    def test_enable_edit_button_when_content_is_locked(self):
        from cms.models import Page
        from django.apps import apps

        user = self.get_superuser()
        version = PageVersionFactory(created_by=user)

        toolbar = get_toolbar(version.content, user, content_mode=True)
        toolbar.post_template_populate()
        edit_button = find_toolbar_buttons("Edit", toolbar.toolbar)[0]

        self.assertEqual(edit_button.name, "Edit")

        cms_extension = apps.get_app_config("djangocms_versioning").cms_extension
        versionable = cms_extension.versionables_by_grouper[Page]
        admin_url = self.get_admin_url(
            versionable.version_model_proxy, "edit_redirect", version.pk
        )
        self.assertEqual(edit_button.url, admin_url)
        self.assertFalse(edit_button.disabled)
        self.assertListEqual(
            edit_button.extra_classes,
            ["cms-btn-action", "cms-form-post-method", "cms-versioning-js-edit-btn"]
        )

    def test_lock_message_when_content_is_locked(self):
        user = self.get_superuser()
        user.first_name = "Firstname"
        user.last_name = "Lastname"
        user.save()
        user_2 = UserFactory(
            is_staff=True,
            is_superuser=True,
            username="admin2",
            email="admin2@123.com",
        )
        version = PageVersionFactory(created_by=user, locked_by=user)

        toolbar = get_toolbar(version.content, user_2, content_mode=True)
        toolbar.post_template_populate()

        for item in toolbar.toolbar.get_right_items():
            if isinstance(item, TemplateItem) and item.template == "djangocms_versioning/admin/lock_indicator.html":
                self.assertEqual(version.locked_message(), f"Locked by {user}")
                break
        else:
            self.assertFalse("locking message not found")

    def test_edit_button_when_content_is_locked_users_username_used(self):
        user = self.get_superuser()
        user.first_name = ""
        user.last_name = ""
        user.save()
        user_2 = UserFactory(
            is_staff=True,
            is_superuser=True,
            username="admin2",
            email="admin2@123.com",
        )
        version = PageVersionFactory(created_by=user, locked_by=user)

        toolbar = get_toolbar(version.content, user_2, content_mode=True)
        toolbar.post_template_populate()
        btn_name = "Unlock"
        unlock_buttons = find_toolbar_buttons(btn_name, toolbar.toolbar)

        self.assertEqual(len(unlock_buttons), 1)


class IntegrationTestCase(CMSTestCase):

    def setUp(self) -> None:
        self.user = self.get_superuser()
        self.version = factories.PollVersionFactory(created_by=self.user, locked_by=self.user)
        self.versionable = PollsCMSConfig.versioning[0]

    def test_unlock_view_with_locking_disabled(self):
        """Tests that unlock view returns 404 if locking is disabled"""
        unlock_url = self.get_admin_url(self.versionable.version_model_proxy, "unlock", self.version.pk)

        with self.login_user_context(self.user):
            conf.LOCK_VERSIONS = False
            response = self.client.post(unlock_url)

        self.assertEqual(response.status_code, 404)



from cms import __version__ as CMS_VERSION
from django.conf import settings

ENABLE_MENU_REGISTRATION = getattr(
    settings, "DJANGOCMS_VERSIONING_ENABLE_MENU_REGISTRATION", CMS_VERSION <= "4.1.0"
)

USERNAME_FIELD = getattr(
    settings, "DJANGOCMS_VERSIONING_USERNAME_FIELD", "username"
)

DEFAULT_USER = getattr(
    settings, "DJANGOCMS_VERSIONING_DEFAULT_USER", None
)

ALLOW_DELETING_VERSIONS = getattr(
    settings, "DJANGOCMS_VERSIONING_ALLOW_DELETING_VERSIONS", False
)

LOCK_VERSIONS = getattr(
    settings, "DJANGOCMS_VERSIONING_LOCK_VERSIONS", False,
)

VERBOSE = getattr(
    settings, "DJANGOCMS_VERSIONING_VERBOSE", True,
)

EMAIL_NOTIFICATIONS_FAIL_SILENTLY = getattr(
    settings, "EMAIL_NOTIFICATIONS_FAIL_SILENTLY", False
)

ON_PUBLISH_REDIRECT = getattr(
    settings, "DJANGOCMS_VERISONING_ON_PUBLISH_REDIRECT", "published"
)
#: Allowed values: "versions", "published", "preview"

VERBOSE_UI = getattr(
    settings, "DJANGOCMS_VERSIONING_VERBOSE_UI", True
)
#: If True, the version admin will be offered in the admin index
#: for each registered versionable model.



from cms.test_utils.testcases import CMSTestCase

from djangocms_versioning import constants
from djangocms_versioning.test_utils import factories
from djangocms_versioning.test_utils.polls.models import PollContent


class TestLatestContentCurrentContent(CMSTestCase):
    def setUp(self):
        poll1 = factories.PollFactory()
        factories.PollVersionFactory(state=constants.PUBLISHED, content__language="de")

        factories.PollVersionFactory(state=constants.ARCHIVED, content__poll=poll1, content__language="de")
        v1 = factories.PollVersionFactory(state=constants.UNPUBLISHED, content__poll=poll1, content__language="de")
        v2 = factories.PollVersionFactory(state=constants.ARCHIVED, content__poll=poll1, content__language="en")
        v3 = factories.PollVersionFactory(state=constants.DRAFT, content__poll=poll1, content__language="en")
        v4 = factories.PollVersionFactory(state=constants.UNPUBLISHED, content__poll=poll1, content__language="fr")

        self.poll = poll1
        self.poll_content1 = v1.content
        self.poll_content2 = v2.content
        self.poll_content3 = v3.content
        self.poll_content4 = v4.content

    def test_latest_content(self):
        """only one version per grouper and grouping field (language) returned."""
        latest_content = PollContent.admin_manager.latest_content(poll=self.poll)
        self.assertEqual(latest_content.count(), 3)
        self.assertIn(self.poll_content1, latest_content)
        self.assertIn(self.poll_content3, latest_content)
        self.assertIn(self.poll_content4, latest_content)

    def test_latest_content_by_language(self):
        """only one version per grouper and grouping field (language) returned. Additional
        filter before or after latest_content() should **not** affect the result."""

        latest_content = PollContent.admin_manager.latest_content().filter(poll=self.poll, language="en")
        self.assertEqual(latest_content.count(), 1)
        self.assertIn(self.poll_content3, latest_content)

        latest_content = PollContent.admin_manager.filter(poll=self.poll, language="en").latest_content()
        self.assertEqual(latest_content.count(), 1)
        self.assertIn(self.poll_content3, latest_content)

        latest_content = PollContent.admin_manager.latest_content().filter(poll=self.poll, language="de")
        self.assertEqual(latest_content.count(), 1)
        self.assertIn(self.poll_content1, latest_content)

        latest_content = PollContent.admin_manager.filter(poll=self.poll, language="de").latest_content()
        self.assertEqual(latest_content.count(), 1)
        self.assertIn(self.poll_content1, latest_content)




from cms.test_utils.testcases import CMSTestCase

from djangocms_versioning import versionables


class VersionableTestCase(CMSTestCase):
    def test_exists_functions_for_models(self):
        """With the example of the poll app test if versionables exists for models"""
        from djangocms_versioning.test_utils.polls.models import (
            Poll,
            PollContent,
        )

        # Check existence
        self.assertTrue(versionables.exists_for_grouper(Poll))
        self.assertTrue(versionables.exists_for_content(PollContent))

        # Check absence
        self.assertFalse(versionables.exists_for_grouper(PollContent))
        self.assertFalse(versionables.exists_for_content(Poll))

    def test_exists_functions_for_objects(self):
        """With the example of the poll app test if versionables exists for objects"""
        from djangocms_versioning.test_utils.factories import (
            PollContentFactory,
            PollFactory,
        )

        poll = PollFactory()
        poll_content = PollContentFactory(poll=poll)

        # Check existence
        self.assertTrue(versionables.exists_for_grouper(poll))
        self.assertTrue(versionables.exists_for_content(poll_content))

        # Check absence
        self.assertFalse(versionables.exists_for_grouper(poll_content))
        self.assertFalse(versionables.exists_for_content(poll))

    def test_get_versionable(self):
        """With the example of the poll app test if versionables for grouper and content models are the same.
        The versionable correctly identfies the content model."""
        from djangocms_versioning.test_utils.polls.models import (
            Poll,
            PollContent,
        )

        v1 = versionables.for_grouper(Poll)
        v2 = versionables.for_content(PollContent)

        self.assertEqual(v1, v2)  # Those are supposed to return the same versionable
        self.assertEqual(v1.content_model, PollContent)  # PollContent should be the content model

    def test_get_versionable_fails_on_unversioned_models(self):
        from djangocms_versioning.test_utils.text.models import Text

        # Versionables do not exists
        self.assertFalse(versionables.exists_for_grouper(Text))
        self.assertFalse(versionables.exists_for_content(Text))

        # Trying to get them raises error
        self.assertRaises(KeyError, lambda: versionables.for_grouper(Text))
        self.assertRaises(KeyError, lambda: versionables.for_content(Text))



from cms.models import PageContent, PageUrl
from cms.test_utils.testcases import CMSTestCase
from django import forms

from djangocms_versioning.forms import grouper_form_factory
from djangocms_versioning.test_utils import factories
from djangocms_versioning.test_utils.polls.models import PollContent


class GrouperFormTestCase(CMSTestCase):
    def tearDown(self):
        grouper_form_factory.cache_clear()

    def test_factory(self):
        pv = factories.PollVersionFactory()
        form_class = grouper_form_factory(PollContent, language="en")

        self.assertIn(forms.Form, form_class.mro())
        self.assertEqual(form_class.__name__, "PollContentGrouperForm")
        self.assertIn("poll", form_class.base_fields)
        self.assertIn(
            (pv.content.poll.pk, str(pv.content.poll)),
            form_class.base_fields["poll"].choices,
        )

    def test_factory_cache(self):
        """Test that grouper_form_factory is cached
        and return value is created once."""
        self.assertEqual(
            id(grouper_form_factory(PollContent, language="en")),
            id(grouper_form_factory(PollContent, language="en")),
        )

    def test_grouper_selector_default_label(self):
        """
        Grouper selector shows the default label format when no override is set
        """
        version = factories.PollVersionFactory()
        form_class = grouper_form_factory(PollContent, language="en")

        self.assertIn(
            (version.content.poll.pk, str(version.content.poll)),
            form_class.base_fields["poll"].choices,
        )

    def test_grouper_selector_non_default_label_unpublished(self):
        """
        Grouper selector shows the PageContent label format when PageContent is set
        """
        version = factories.PageVersionFactory()
        form_class = grouper_form_factory(PageContent, version.content.language)
        label = "No available title (Unpublished)"
        self.assertIn(
            (version.content.page.pk, label), form_class.base_fields["page"].choices
        )

    def test_grouper_selector_non_default_label(self):
        """
        Grouper selector shows the PageContent label format when PageContent is set
        """
        version = factories.PageVersionFactory()
        PageUrl.objects.create(
            page=version.content.page,
            language=version.content.language,
            path="test",
            slug="test",
        )
        version.publish(version.created_by)
        form_class = grouper_form_factory(PageContent, version.content.language)
        label = f"{version.content.title} (/{version.content.page.get_path(version.content.language)}/)"
        self.assertIn(
            (version.content.page.pk, label), form_class.base_fields["page"].choices
        )



import warnings
from copy import copy

from django.contrib.auth import get_user_model
from django.db import models

from . import constants
from .constants import PUBLISHED
from .models import Version


class PublishedContentManagerMixin:
    """Manager mixin used for overriding the managers of content models"""
    versioning_enabled = True

    def get_queryset(self):
        """Limit query to published content
        """
        queryset = super().get_queryset()
        if not self.versioning_enabled:
            return queryset
        return queryset.filter(versions__state=PUBLISHED)

    def create(self, *args, **kwargs):
        obj = super().create(*args, **kwargs)
        created_by = kwargs.get("created_by", None)
        if not isinstance(created_by, get_user_model()):
            created_by = getattr(self, "_user", None)
        if created_by:
            Version.objects.create(content=obj, created_by=created_by)
        else:
            warnings.warn(
                f"No user has been supplied when creating a new {obj.__class__.__name__} object. "
                f"No version could be created. Make sure that the creating code also creates a "
                f"Version objects or use {obj.__class__.__name__}.objects.with_user(user).create(...)",
                UserWarning, stacklevel=2,
            )
        return obj

    def with_user(self, user):
        if not isinstance(user, get_user_model()) and user is not None:
            import inspect

            curframe = inspect.currentframe()
            callframe = inspect.getouterframes(curframe, 2)
            calling_function = callframe[1][3]
            raise ValueError(
                    f"With versioning enabled, {calling_function} requires a {get_user_model().__name__} instance "
                    f"to be passed as created_by argument"
                )
        new_manager = copy(self)
        new_manager._user = user
        return new_manager


class AdminQuerySetMixin:
    # Annotation for latest pk of draft or published version
    _DraftOrPublished = models.Max(
        models.Case(
            models.When(versions__state__in=(constants.DRAFT, constants.PUBLISHED),
                        then="versions__pk"),
            default=models.Value(0),
        )
    )

    # Annotation for latest pk of any other version
    _AnyOther = models.Max(
        models.Case(
            models.When(
                ~models.Q(versions__state__in=(constants.DRAFT, constants.PUBLISHED)),
                then="versions__pk"),
            default=models.Value(0),
        )
    )

    def _chain(self):
        # Also clone group by key when chaining querysets!
        clone = super()._chain()
        clone._group_by_key = self._group_by_key
        return clone

    def current_content(self, **kwargs):
        """Returns a queryset current content versions. Current versions are either draft
        versions or published versions (in that order). This optimized query assumes that
        draft versions always have a higher pk than any other version type. This is true as long as
        no other version type can be converted to draft without creating a new version."""

        pk_filter = self.filter(versions__state__in=(constants.DRAFT, constants.PUBLISHED))\
            .values(*self._group_by_key)\
            .annotate(vers_pk=models.Max("versions__pk"))\
            .values("vers_pk")
        return self.filter(versions__pk__in=pk_filter, **kwargs)

    def latest_content(self, **kwargs):
        """Returns the "latest" content object which is in this order
           1. a draft version (should it exist)
           2. a published version (should it exist)
           3. any other version with the highest pk

        This filter assumes that there can only be one draft created and that the draft as
        the highest pk of all versions (should it exist).
        """

        latest = (self.values(*self._group_by_key)
                  .annotate(h1=self._DraftOrPublished, h2=self._AnyOther)
                  .annotate(vers_pk=models.Case(models.When(h1__gt=0, then="h1"), default="h2"))
                  .values("vers_pk")
                  )
        return self.filter(versions__pk__in=latest, **kwargs)


class AdminManagerMixin:
    versioning_enabled = True
    _group_by_key = []

    def get_queryset(self):
        qs_class = super().get_queryset().__class__
        if not self._group_by_key:
            # Not initialized (e.g. by using content_set(manager="admin_manager"))?
            # Get grouping fields from versionable
            from . import versionables
            versionable = versionables.for_content(self.model)
            self._group_by_key = list(versionable.grouping_fields)
        qs = type(
            f"Admin{qs_class.__name__}",
            (AdminQuerySetMixin, qs_class),
            {"_group_by_key": self._group_by_key}  # Pass grouping fields to queryset
        )(self.model, using=self._db)
        return qs

    def current_content(self, **kwargs):  # pragma: no cover
        """Syntactic sugar: admin_manager.current_content()"""
        return self.get_queryset().current_content(**kwargs)

    def latest_content(self, **kwargs):  # pragma: no cover
        """Syntactic sugar: admin_manager.latest_content()"""
        return self.get_queryset().latest_content(**kwargs)



from cms.extensions import PageContentExtension
from cms.extensions.extension_pool import extension_pool
from django.db import models


class PollPageContentExtension(PageContentExtension):
    votes = models.IntegerField()


extension_pool.register(PollPageContentExtension)



from cms.app_base import CMSAppConfig

from djangocms_versioning.datastructures import VersionableItem, default_copy

from .models import BlogContent, CommentContent


class BlogpostCMSConfig(CMSAppConfig):
    djangocms_versioning_enabled = True
    versioning = [
        VersionableItem(
            content_model=BlogContent,
            grouper_field_name="blogpost",
            copy_function=default_copy,
        ),
        VersionableItem(
            content_model=CommentContent,
            grouper_field_name="comment",
            copy_function=default_copy,
        ),
    ]



from setuptools import find_packages, setup

import djangocms_versioning

INSTALL_REQUIREMENTS = [
    "Django>=3.2",
    "django-cms>=4.1.1",
    "django-fsm<3",
    "packaging",
]

setup(
    name="djangocms-versioning",
    packages=find_packages(),
    include_package_data=True,
    version=djangocms_versioning.__version__,
    description=djangocms_versioning.__doc__,
    long_description=open("README.rst").read(),
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
    ],
    install_requires=INSTALL_REQUIREMENTS,
    author="Divio AG",
    test_suite="test_settings.run",
    author_email="info@divio.ch",
    maintainer="Django CMS Association and contributors",
    maintainer_email="info@django-cms.org",
    url="https://github.com/django-cms/djangocms-versioning",
    license="BSD",
)



from urllib.parse import parse_qs, urlparse

from cms.test_utils.testcases import CMSTestCase

from djangocms_versioning.helpers import (
    version_list_url,
    version_list_url_for_grouper,
)
from djangocms_versioning.test_utils import factories


class VersionListUrlsTestCase(CMSTestCase):
    def test_version_list_url(self):
        pv = factories.PollVersionFactory(content__language="en")
        url = version_list_url(pv.content)
        parsed = urlparse(url)
        self.assertEqual(
            parsed.path, "/en/admin/djangocms_versioning/pollcontentversion/"
        )
        self.assertEqual(
            {k: v[0] for k, v in parse_qs(parsed.query).items()},
            {"poll": str(pv.grouper.pk), "language": "en"},
        )

    def test_version_list_url_for_grouper(self):
        pv = factories.PollVersionFactory()
        self.assertEqual(
            version_list_url_for_grouper(pv.grouper),
            f"/en/admin/djangocms_versioning/pollcontentversion/?poll={pv.grouper.pk}",
        )



from cms import constants as cms_constants
from cms.cms_menus import CMSMenu as OriginalCMSMenu
from cms.test_utils.testcases import CMSTestCase
from cms.toolbar.toolbar import CMSToolbar
from cms.toolbar.utils import get_object_preview_url
from django.contrib.auth.models import AnonymousUser
from django.template import Context, Template
from django.test import RequestFactory
from django.test.utils import override_settings
from menus.menu_pool import menu_pool

from djangocms_versioning.cms_menus import CMSMenu
from djangocms_versioning.test_utils.factories import (
    PageVersionFactory,
    UserFactory,
)


class CMSVersionedMenuTestCase(CMSTestCase):
    def setUp(self):
        super().setUp()
        from djangocms_versioning.test_utils.factories import TreeNode

        def get_page(title, path, parent=None):
            return {
                "content__title": title,
                "content__menu_title": "",
                "content__in_navigation":  True,
                "content__limit_visibility_in_menu":  None,
                "content__language": "en",
                "content__page__node__path" if TreeNode else "content__page__path": path,
                "content__page__node__parent" if TreeNode else "content__page__parent": parent,
            }
        self._page_1 = PageVersionFactory(**get_page("page_content_1", "0001"))
        self._page_2 = PageVersionFactory(**get_page("page_content_2", "0002"))
        self._page_2_1 = PageVersionFactory(**get_page(
            "page_content_2_1",
            "00020001",
            self._page_2.content.page.node if TreeNode else self._page_2.content.page,
        ))
        self._page_2_2 = PageVersionFactory(**get_page(
            "page_content_2_2",
            "00020002",
            self._page_2.content.page.node if TreeNode else self._page_2.content.page,
        ))
        self._page_3 = PageVersionFactory(**get_page("page_content_3", "0003"))

    def _render_menu(self, user=None, **kwargs):
        request = RequestFactory().get("/")

        if not user:
            is_auth_user = kwargs.get("is_auth_user", True)
            user = self.get_superuser() if is_auth_user else AnonymousUser()

        request.user = user
        request.session = {}
        toolbar = CMSToolbar(request)

        if kwargs.get("edit_mode", False):
            toolbar.edit_mode_active = True
            toolbar.preview_mode_active = False
        elif kwargs.get("preview_mode", False):
            toolbar.edit_mode_active = False
            toolbar.preview_mode_active = True
        else:
            toolbar.edit_mode_active = False
            toolbar.preview_mode_active = False

        request.toolbar = toolbar
        context = {"request": request}
        template = Template("{% load menu_tags %}" "{% show_menu 0 100 100 100 %}")
        template.render(Context(context))
        return context

    def _assert_node(self, node, version, edit_or_preview=True):
        content = version.content
        self.assertEqual(node.title, content.title)

        if edit_or_preview:
            self.assertEqual(node.url, get_object_preview_url(content))
        else:
            self.assertEqual(node.url, content.get_absolute_url())

    def test_core_cms_menu_is_removed(self):
        menu_pool.discover_menus()
        registered_menus = menu_pool.get_registered_menus(for_rendering=True)
        self.assertNotIn(OriginalCMSMenu, registered_menus.values())
        self.assertIn(CMSMenu, registered_menus.values())

    def test_no_menu_if_no_published_pages_in_public_mode(self):
        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 0)

    def test_show_menu_with_draft_pages_in_edit_mode(self):
        context = self._render_menu(edit_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 3)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], self._page_2)
        children = nodes[1].children
        self.assertEqual(len(children), 2)
        self._assert_node(children[0], self._page_2_1)
        self._assert_node(children[1], self._page_2_2)
        self._assert_node(nodes[2], self._page_3)

    def test_show_menu_with_draft_pages_in_preview_mode(self):
        context = self._render_menu(edit_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 3)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], self._page_2)
        children = nodes[1].children
        self.assertEqual(len(children), 2)
        self._assert_node(children[0], self._page_2_1)
        self._assert_node(children[1], self._page_2_2)
        self._assert_node(nodes[2], self._page_3)

    def test_show_menu_with_published_nodes_only_in_public_mode(self):
        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 3)
        self._assert_node(nodes[0], self._page_1)

        # Publish Page 1.
        self._page_1.publish(self.get_superuser())

        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 1)
        self._assert_node(nodes[0], self._page_1, False)

    def test_not_show_published_child_node_if_parent_is_draft_in_public_mode(self):
        # Publish Page 1.
        self._page_1.publish(self.get_superuser())

        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 1)
        self._assert_node(nodes[0], self._page_1, False)

        # Publish Page 2_1. This should not be rendered because
        # Page 2 is not published.
        self._page_2_1.publish(self.get_superuser())

        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 1)
        self._assert_node(nodes[0], self._page_1, False)

    def test_show_child_node_if_parent_is_published_in_public_mode(self):
        # Publish Page 2 and Page 2_1.
        self._page_2.publish(self.get_superuser())
        self._page_2_1.publish(self.get_superuser())

        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 1)
        self._assert_node(nodes[0], self._page_2, False)
        children = nodes[0].children
        self.assertEqual(len(children), 1)
        self._assert_node(children[0], self._page_2_1, False)

        # Publish Page 2_1.
        self._page_2_2.publish(self.get_superuser())

        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 1)
        self._assert_node(nodes[0], self._page_2, False)
        children = nodes[0].children
        self.assertEqual(len(children), 2)
        self._assert_node(children[0], self._page_2_1, False)
        self._assert_node(children[1], self._page_2_2, False)

    def test_not_show_child_node_if_not_in_navigation(self):
        # Page 2_1 hidden in navigation.
        self._page_2_1.content.in_navigation = False
        self._page_2_1.content.save()

        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 3)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], self._page_2)
        children = nodes[1].children
        self.assertEqual(len(children), 1)
        self._assert_node(children[0], self._page_2_2)
        self._assert_node(nodes[2], self._page_3)

    def test_not_show_nodes_if_hidden_in_navigation(self):
        # Hide Page 1 and Page 2_1 in navigation.
        self._page_1.content.in_navigation = False
        self._page_1.content.save()
        self._page_2_1.content.in_navigation = False
        self._page_2_1.content.save()

        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 2)
        self._assert_node(nodes[0], self._page_2)
        children = nodes[0].children
        self.assertEqual(len(children), 1)
        self._assert_node(children[0], self._page_2_2)
        self._assert_node(nodes[1], self._page_3)

    def test_not_show_child_nodes_if_parent_not_in_navigation(self):
        # Page 2 hidden in navigation.
        self._page_2.content.in_navigation = False
        self._page_2.content.save()

        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 2)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], self._page_3)

    def test_show_home_page_children_nodes_even_if_home_page_is_hidden_in_navigation(
        self
    ):
        # Make Page 2 home and hide in navigation.
        self._page_2.content.page.is_home = True
        self._page_2.content.page.save()
        self._page_2.content.in_navigation = False
        self._page_2.content.save()

        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 4)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], self._page_2_1)
        self._assert_node(nodes[2], self._page_2_2)
        self._assert_node(nodes[3], self._page_3)

    def test_show_published_nodes_if_draft_not_exist_in_non_public_mode(self):
        # Publish Page 2 and Page 2_2.
        self._page_2.publish(self.get_superuser())
        self._page_2_2.publish(self.get_superuser())

        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 3)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], self._page_2)
        children = nodes[1].children
        self.assertEqual(len(children), 2)
        self._assert_node(children[0], self._page_2_1)
        self._assert_node(children[1], self._page_2_2)
        self._assert_node(nodes[2], self._page_3)

    def test_show_draft_nodes_if_draft_exist_in_non_public_mode(self):
        # Publish Page 2 and Page 2_2.
        self._page_2.publish(self.get_superuser())
        self._page_2_2.publish(self.get_superuser())
        # Create new drafts for both Page 2 and Page 2_2.
        _page_2_new_draft = self._page_2.copy(self.get_superuser())
        _page_2_2_new_draft = self._page_2_2.copy(self.get_superuser())
        # Make some changes to the new drafts (Just to verify).
        _page_2_new_draft.content.title = "page_content_2_new_draft"
        _page_2_new_draft.content.save()
        _page_2_2_new_draft.content.title = "page_content_2_2_new_draft"
        _page_2_2_new_draft.content.save()

        context = self._render_menu(preview_mode=True)
        nodes = context["children"]
        self.assertEqual(len(nodes), 3)
        self._assert_node(nodes[0], self._page_1)
        self._assert_node(nodes[1], _page_2_new_draft)
        children = nodes[1].children
        self.assertEqual(len(children), 2)
        self._assert_node(children[0], self._page_2_1)
        self._assert_node(children[1], _page_2_2_new_draft)
        self._assert_node(nodes[2], self._page_3)

    def test_attr_set_properly_to_node(self):
        # To test the attr.limit_visibility_in_menu we will set
        # to cms_constants.VISIBILITY_USERS, so anonymous users
        # cannot see the menu node for Page 1.
        self._page_1.content.limit_visibility_in_menu = cms_constants.VISIBILITY_USERS
        self._page_1.content.save()
        self._page_1.publish(self.get_superuser())

        # Test for anonymous user.
        context = self._render_menu(is_auth_user=False)
        nodes = context["children"]
        self.assertEqual(len(nodes), 0)

        # Test for logged in user.
        context = self._render_menu()
        nodes = context["children"]
        self.assertEqual(len(nodes), 1)
        self._assert_node(nodes[0], self._page_1, False)

    @override_settings(CMS_PUBLIC_FOR="staff")
    def test_show_menu_only_visible_for_user(self):
        from cms.models import ACCESS_PAGE, PagePermission
        from django.contrib.auth.models import Group

        group = Group.objects.create(name="test_group")
        user = UserFactory()
        user.groups.add(group)
        q_args = {"grant_on": ACCESS_PAGE, "group": group}
        # Restrict pages for the user so that we can test
        # whether the can_view=True only pages are visible in the menu.
        PagePermission.objects.create(
            can_view=False, page=self._page_1.content.page, **q_args
        )
        PagePermission.objects.create(
            can_view=True, page=self._page_2.content.page, **q_args
        )
        PagePermission.objects.create(
            can_view=False, page=self._page_2_1.content.page, **q_args
        )
        PagePermission.objects.create(
            can_view=True, page=self._page_2_2.content.page, **q_args
        )
        PagePermission.objects.create(
            can_view=True, page=self._page_3.content.page, **q_args
        )

        context = self._render_menu(user=user, preview_mode=True)
        nodes = context["children"]
        # At this point, only Page 2, Page 2_2 and Page 3 should
        # be rendered in the menu.
        self.assertEqual(len(nodes), 2)
        self._assert_node(nodes[0], self._page_2)
        children = nodes[0].children
        self.assertEqual(len(children), 1)
        self._assert_node(children[0], self._page_2_2)
        self._assert_node(nodes[1], self._page_3)



from unittest.mock import patch

from cms.models.contentmodels import PageContent
from cms.test_utils.testcases import CMSTestCase
from django.db import models

from djangocms_versioning import constants, helpers
from djangocms_versioning.helpers import replace_manager
from djangocms_versioning.managers import (
    AdminManagerMixin,
    PublishedContentManagerMixin,
)
from djangocms_versioning.test_utils.factories import (
    PageFactory,
    PageVersionFactory,
)
from djangocms_versioning.test_utils.people.models import Person


class ContentModelTestCase(CMSTestCase):
    def tearDown(self):
        Person.add_to_class("objects", models.Manager())

    def test_replace_default_manager(self):
        self.assertNotIn(PublishedContentManagerMixin, Person.objects.__class__.mro())
        replace_manager(Person, "objects", PublishedContentManagerMixin)
        self.assertIn(PublishedContentManagerMixin, Person.objects.__class__.mro())

        self.assertFalse(hasattr(Person, "admin_manager"))
        replace_manager(Person, "admin_manager", AdminManagerMixin)
        self.assertIn(AdminManagerMixin, Person.admin_manager.__class__.mro())

    def test_replace_default_manager_twice(self):
        replace_manager(Person, "objects", PublishedContentManagerMixin)
        with patch.object(helpers, "manager_factory") as mock:
            replace_manager(Person, "objects", PublishedContentManagerMixin)
        mock.assert_not_called()

        original_manager = Person._original_manager

        replace_manager(Person, "admin_manager", AdminManagerMixin)
        with patch.object(helpers, "manager_factory") as mock:
            replace_manager(Person, "admin_manager", AdminManagerMixin)
        mock.assert_not_called()

        # Replacing admin_manager did not overwrite _original_manager?
        self.assertEqual(Person._original_manager, original_manager)


class AdminManagerTestCase(CMSTestCase):
    def create_page_content(self, page, language, version_state):
        version = PageVersionFactory(content__page=page, content__language=language)
        if version_state == constants.PUBLISHED:
            version.publish(self.get_superuser())
        elif version_state == constants.ARCHIVED:
            version.archive(self.get_superuser())

    def setUp(self) -> None:
        self.pages1 = [PageFactory() for i in range(2)]
        for page in self.pages1:
            self.create_page_content(page, "en", constants.DRAFT)
            self.create_page_content(page, "it", constants.PUBLISHED)

        self.pages2 = [PageFactory() for i in range(2)]
        for page in self.pages2:
            self.create_page_content(page, "en", constants.PUBLISHED)
            self.create_page_content(page, "en", constants.DRAFT)
            self.create_page_content(page, "it", constants.ARCHIVED)
            self.create_page_content(page, "it", constants.PUBLISHED)

    def test_current_content(self):
        # 12 PageContent versions in total
        self.assertEqual(len(list(
            PageContent.admin_manager.all()
        )), 12)

        # 4 total PageContent versions for self.pages1 (2 pages x 2 languages)
        qs = PageContent.admin_manager.filter(page__in=self.pages1)
        self.assertEqual(len(qs), 4)
        self.assertEqual(qs._group_by_key, ["page", "language"])
        self.assertEqual(len(list(
            PageContent.admin_manager.filter(page__in=self.pages1).current_content()
        )), 4, f"{list(PageContent.admin_manager.filter(page__in=self.pages1).current_content())}")
        # 2 current PageContent versions for self.pages2
        self.assertEqual(len(list(
            PageContent.admin_manager.filter(page__in=self.pages2).current_content()
        )), 4)

        # Now unpublish all published in pages2
        for page in self.pages2:
            for content in page.pagecontent_set.all():
                content.versions.first().unpublish(self.get_superuser())

        # 2 current PageContent versions for self.pages2
        self.assertEqual(len(list(
            PageContent.admin_manager.filter(page__in=self.pages2).current_content()
        )), 2)



from __future__ import unicode_literals

from django.db import connection, migrations


def _alter_unique_together(schema_editor, model, old_unique_together, new_unique_together):
    if connection.settings_dict["ENGINE"].rsplit(".")[-1] == "mysql":
        # Switch off atomic for mysql
        in_atomic_block = schema_editor.connection.in_atomic_block
        schema_editor.connection.in_atomic_block = False
        try:
            schema_editor.alter_unique_together(
                model, old_unique_together, new_unique_together
            )
        finally:
            schema_editor.connection.in_atomic_block = in_atomic_block
    else:
        schema_editor.alter_unique_together(
            model, old_unique_together, new_unique_together
        )


def forwards(apps, schema_editor):
    PageContent = apps.get_model("cms", "PageContent")
    old_unique_together = PageContent._meta.unique_together
    new_unique_together = old_unique_together - {("language", "page")}
    _alter_unique_together(schema_editor, PageContent, old_unique_together, new_unique_together)


def backwards(apps, schema_editor):
    PageContent = apps.get_model("cms", "PageContent")
    old_unique_together = PageContent._meta.unique_together
    new_unique_together = old_unique_together | {("language", "page")}
    _alter_unique_together(schema_editor, PageContent, old_unique_together, new_unique_together)


class Migration(migrations.Migration):

    dependencies = [
        ("djangocms_versioning", "0008_auto_20180820_1754"),
        ("cms", "0032_remove_title_to_pagecontent"),
    ]

    operations = [migrations.RunPython(forwards, backwards)]
